functor Prop
  (structure Ctx' : CTX
   structure Term' : TERM
     sharing Term'.Ctx = Ctx')
  : PROP =
struct

  structure Ctx = Ctx'
  structure Term = Term'

  local
    open Ctx
    infix $
  in
					(* Propositions   *)
    datatype prop =			(* A ::=          *)
      Atom of string * Term.term list	(*   P(t1,...,tn) *)
    | Tensor of prop * prop		(* | A1 * A2      *)
    | One				(* | 1            *)
    | Lolli of prop * prop		(* | A1 -o A2     *)
    | With of prop * prop		(* | A1 & A2      *)
    | Top				(* | T            *)
    | Plus of prop * prop		(* | A1 + A2      *)
    | Zero				(* | 0            *)
    | Implies of prop * prop		(* | A1 => A2     *)
    | Bang of prop			(* | ! A          *)
    | Forall of Term.var * prop		(* | forall x. A  *)
    | Exists of Term.var * prop		(* | exists x. A  *)

    (* subst (t,x) A ===> [t/x] A; t closed *)
    fun subst (t,x) =
	let
	  fun sbst (Atom (p, ss)) = Atom (p, map (Term.subst (t,x)) ss)
	    | sbst (Tensor (A1, A2)) = Tensor (sbst A1, sbst A2)
	    | sbst (One) = One
	    | sbst (Lolli (A1, A2)) = Lolli (sbst A1, sbst A2)
	    | sbst (With (A1, A2)) = With (sbst A1, sbst A2)
	    | sbst (Top) = Top
	    | sbst (Plus (A1, A2)) = Plus (sbst A1, sbst A2)
	    | sbst (Zero) = Zero
	    | sbst (Implies (A1, A2)) = Implies (sbst A1, sbst A2)
	    | sbst (Bang (A1)) = Bang (sbst A1)
	    | sbst (A as Forall (y, A1)) =
	      if x = y then A else Forall (y, sbst A1)
	    | sbst (A as Exists (y, A1)) =
	      if x = y then A else Exists (y, sbst A1)
	in
	  sbst
	end

    (* eq (A, B) ===> true  iff  A = B; A and B closed *)
    (*
       A = B  if A and B are equal modulo renaming of bound variables.
       For example, forall x. P(x) = forall y. P(y),
       but not P('a) = P('b) for distinct parameters a and b.
    *)
    fun eq (Atom (p, ts), Atom (q, ss)) =
	  (p = q) andalso Term.eq' (ts, ss)
      | eq (Tensor (A1, A2), Tensor (B1, B2)) =
	  eq (A1, B1) andalso eq (A2, B2)
      | eq (One, One) = true
      | eq (Lolli (A1, A2), Lolli (B1, B2)) =
	  eq (A1, B1) andalso eq (A2, B2)
      | eq (With (A1, A2), With (B1, B2)) =
	  eq (A1, B1) andalso eq (A2, B2)
      | eq (Top, Top) = true
      | eq (Plus (A1, A2), Plus (B1, B2)) =
	  eq (A1, B1) andalso eq (A2, B2)
      | eq (Zero, Zero) = true
      | eq (Implies (A1, A2), Implies (B1, B2)) =
	  eq (A1, B1) andalso eq (A2, B2)
      | eq (Bang (A1), Bang (A2)) =
	  eq (A1, A2)
      | eq (Forall (x, A1), Forall (y, B1)) =
	let
	  (* a will be different from any parameter in A1 or B1 *)
	  val a = Term.Parm (Term.newParm ())
	in
	  eq (subst (a,x) A1, subst (a,x) B1)
	end
      | eq (Exists (x, A1), Exists (y, B1)) =
	let
	  (* a will be different from any parameter in A1 or B1 *)
	  val a = Term.Parm (Term.newParm ())
	in
	  eq (subst (a,x) A1, subst (a,x) B1)
	end
      | eq _ = false

    (* closed P A ===> true  iff  A closed wrt P *)
    fun closed P =
	let fun cld (Atom (p, ts)) = List.all (Term.closed P) ts
	      | cld (Tensor (A1, A2)) = cld A1 andalso cld A2
	      | cld (One) = true
	      | cld (Lolli (A1, A2)) = cld A1 andalso cld A2
	      | cld (With (A1, A2)) = cld A1 andalso cld A2
	      | cld (Top) = true
	      | cld (Plus (A1, A2)) = cld A1 andalso cld A2
	      | cld (Zero) = true
	      | cld (Implies (A1, A2)) = cld A1 andalso cld A2
	      | cld (Bang (A1)) = cld A1
	      | cld (Forall (x, A1)) =
		let
		  val a = Term.newParm ()
		in 
		   closed (P $ a) (subst (Term.Parm (a), x) A1)
		end
	      | cld (Exists (x, A1)) =
		let
		  val a = Term.newParm ()
		in
		   closed (P $ a) (subst (Term.Parm (a), x) A1)
		end
	in
	  cld
	end 

    fun toStringTs (t::nil) = Term.toString t
      | toStringTs (t::ts) = Term.toString t ^ "," ^ toStringTs ts

    fun parens (true, string) = "(" ^ string ^ ")"
      | parens (false, string) = string

    (* toStringB (b, A) = string, where b = true if we need to
       parenthesize binary operators or quantifiers.
       This will still insert numerous quantifiers which would
       be unnecessary by standard conventions, but it helps
       in avoiding misinterpretations.
    *)
    fun toStringB (b, Atom(p, nil)) = p
      | toStringB (b, Atom(p, ts)) = p ^ "(" ^ toStringTs ts ^ ")"
      | toStringB (b, Tensor(A1, A2)) = binop (b, "*", (A1, A2))
      | toStringB (b, One) = "1"
      | toStringB (b, Lolli(A1, A2)) = binop (b, "-o", (A1, A2))
      | toStringB (b, With(A1, A2)) = binop (b, "&", (A1, A2))
      | toStringB (b, Top) = "T"
      | toStringB (b, Plus(A1, A2)) = binop (b, "+", (A1, A2))
      | toStringB (b, Zero) = "0"
      | toStringB (b, Implies(A1, A2)) = binop (b, "=>", (A1, A2))
      | toStringB (b, Bang(A1)) = "!" ^ toStringB (true, A1)
      | toStringB (b, Forall(x, A1)) =
        parens (b, "forall" ^ " " ^ x ^ ". " ^ toStringB (false, A1))
      | toStringB (b, Exists(x, A1)) =
	parens (b, "exists" ^ " " ^ x ^ ". " ^ toStringB (false, A1))
    and binop (b, opr, (A1, A2)) =
	  parens (b, toStringB (true, A1) ^ " "
		     ^ opr ^ " " ^ toStringB (true, A2))

    fun toString (A) = toStringB (false, A)

  end  (* local ... *)
end;  (* functor Prop *)
