functor Term
  (structure Ctx' : CTX)
  : TERM =
struct

  structure Ctx = Ctx'

  local
    open Ctx
    infix $
  in

    type const = string			(* Constants c,f  *)
    type parm = string			(* Parameters a   *)
    type var = string			(* Variables x    *)

					(* Terms          *)
    datatype term =			(* t ::=          *)
      Parm of parm			(*   a            *)
    | Var of var			(* | x            *)
    | Root of const * term list		(* | f(t1,...,tn) *)

    (*
       We say "t is closed" if t contains no free variables.
       We say "ts is closed" if all terms in ts are closed.
    *)

    (* subst (t,x) s ===> [t/x] s; t must be closed *)
    fun subst (t, x) s =
	let fun sbst (s as Parm _) = s
	      | sbst (s as Var(y)) =
		if x = y then t else s
	      | sbst (Root (g, ss)) =
		  Root (g, map sbst ss)
	in
	  sbst s
	end

    (* eq (t, s) ===> true  iff  t = s; t and s must be closed *)
    (* eq' (ts, ss) ===> true  iff  ts = ss; ts and ss must be closed *)
    fun eq (Parm(a), Parm(b)) = (a = b)
      | eq (Root(f, ts), Root(g, ss)) =
	  (f = g) andalso eq' (ts, ss)
      | eq _ = false			(* Var(_) not permitted! *)
    and eq' (nil, nil) = true
      | eq' (t::ts, s::ss) = eq (t, s) andalso eq' (ts, ss)
      | eq' _ = false

    (* closed P t ===> true  iff t closed wrt P *)
    (*
       t is closed wrt P if it is closed and
       all parameters in t are declared in parameter context P.
    *)
    fun closed P =
	let fun cld (Parm(a)) = member (a, P)
	      | cld (Var _) = false
	      | cld (Root (f, ts)) = cld' ts
	    and cld' (nil) = true
	      | cld' (t::ts) = cld t andalso cld' ts
	in cld end

    (* legal a ===> true  iff a is a legal external parameter *)
    (*
       External parameters are not allows to start with a digit,
       so they cannot conflict with internally generated parameters.
    *)
    fun legal (x) = not (Char.isDigit (String.sub (x, 0)))

    local
      (* local counter to guarantee new internal parameters *)
      val counter = ref 0
    in
      (* newParm () ===> a,  where a is a new internal parameter *)
      (*
         Internal parameters are natural numbers,
	 so they cannot conflict with external parameters.
      *)
      fun newParm () =
	  (counter := !counter+1;
	   Int.toString (!counter))
    end

    (* Constants c are printed as c() to distinguish them from variables. *)
    (* Parameters a are printed as 'a. *)
    (* Variables x are printed simply as x. *)
    (* Output may look wrong if variables are named 'x. *)
    fun toString (Parm(a)) = "'" ^ a
      | toString (Var(x)) = x
      | toString (Root(f,ts)) = f ^ "(" ^ toString' ts ^ ")"
    and toString' (nil) = ""
      | toString' (t::nil) = toString t
      | toString' (t::ts) = toString t ^ "," ^ toString' ts

  end  (* local ... *)

end;  (* functor Term *)
